import uuid
import enum
import datetime

from sqlalchemy import event
from sqlalchemy.ext.hybrid import hybrid_property

from portal import db
from portal.database.mixins import EntityBase
from portal.database.exceptions import PortalDbError
from portal.tools import time
# uuid as primary key source:
# https://stackoverflow.com/questions/36806403/cant-render-element-of-type-class-sqlalchemy-dialects-postgresql-base-uuid


def _repr(instance):
    result = f"{instance.__class__.__name__}: "
    for key, value in vars(instance).items():
        if not key.startswith("_"):
            result += f"{key}={value} "
    return result


class User(db.Model, EntityBase):
    __tablename__ = 'user'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    uco = db.Column(db.Integer, unique=True, nullable=False)
    email = db.Column(db.String(50), unique=True, nullable=False)
    username = db.Column(db.String(15), unique=True, nullable=False)
    name = db.Column(db.String(50))
    is_admin = db.Column(db.Boolean, default=False)
    password_hash = db.Column(db.String(50), default=None)  # optional - after password change

    submissions = db.relationship("Submission", back_populates="user", cascade="all, delete-orphan",
                                  passive_deletes=True)
    review_items = db.relationship("ReviewItem", back_populates="user")

    @property
    def courses(self):
        result = []
        for role in self.roles:
            if role.course not in result:
                result.append(role.course)
        return result

    def __init__(self, uco, email, username, is_admin=False) -> None:
        self.uco = uco
        self.email = email
        self.username = username
        self.is_admin = is_admin

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Course(db.Model, EntityBase):
    __tablename__ = 'course'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(50), unique=True, nullable=False)
    codename = db.Column(db.String(15), unique=True, nullable=False)
    notes_access_token = db.Column(db.String(100))  # TODO: check length

    # Info on cascades: http://docs.sqlalchemy.org/en/latest/orm/cascades.html
    # Cascades here simulate a composition - roles, groups and projects exist only when associated to a course
    roles = db.relationship("Role", back_populates="course", cascade="all, delete-orphan", passive_deletes=True)
    groups = db.relationship("Group", back_populates="course", cascade="all, delete-orphan", passive_deletes=True)
    projects = db.relationship("Project", back_populates="course", cascade="all, delete-orphan", passive_deletes=True)

    def __init__(self, name, codename) -> None:
        self.name = name
        self.codename = codename

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class ProjectState(enum.Enum):
    ACTIVE = 1
    INACTIVE = 2
    ARCHIVED = 3


class Project(db.Model, EntityBase):
    __tablename__ = 'project'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(50), nullable=False)

    config = db.relationship("ProjectConfig", back_populates="project", uselist=False,
                             cascade="all, delete-orphan", passive_deletes=True)

    course_id = db.Column(db.Integer, db.ForeignKey('course.id'), nullable=False)
    course = db.relationship("Course", back_populates="projects", uselist=False)
    submissions = db.relationship("Submission", back_populates="project", cascade="all, delete-orphan",
                                  passive_deletes=True)

    __table_args__ = (
        db.UniqueConstraint('course_id', 'name', name='course_unique_name'),
    )

    def state(self, timestamp=datetime.datetime.now()) -> ProjectState:
        if self.config.submissions_allowed_from <= timestamp < self.config.submissions_allowed_to:
            return ProjectState.ACTIVE
        elif timestamp >= self.config.archive_from:
            return ProjectState.ARCHIVED
        else:
            return ProjectState.INACTIVE

    def set_config(self, **kwargs):
        for k, w in kwargs.items():
            if k in ('test_files_source', 'file_whitelist', 'pre_submit_script', 'post_submit_script',
                     'submission_parameters', 'submission_scheduler_config', 'submissions_allowed_from',
                     'submissions_allowed_to', 'archive_from'):
                setattr(self.config, k, w)

    def __init__(self, course, name, test_files_source) -> None:
        self.course = course
        self.name = name
        self.config = ProjectConfig(project=self, test_files_source=test_files_source)

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class ProjectConfig(db.Model, EntityBase):
    __tablename__ = 'projectConfig'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    project_id = db.Column(db.Integer, db.ForeignKey('project.id'), nullable=False)
    project = db.relationship("Project", back_populates="config", uselist=False)

    test_files_source = db.Column(db.String(100), nullable=False)  # a git repository URL
    file_whitelist = db.Column(db.Text)
    pre_submit_script = db.Column(db.Text)
    post_submit_script = db.Column(db.Text)
    submission_parameters = db.Column(db.Text)
    submission_scheduler_config = db.Column(db.Text)

    _submissions_allowed_from = db.Column(db.TIMESTAMP(timezone=True))
    _submissions_allowed_to = db.Column(db.TIMESTAMP(timezone=True))
    _archive_from = db.Column(db.TIMESTAMP(timezone=True))

    @hybrid_property
    def submissions_allowed_from(self):
        return time.strip_seconds(self._submissions_allowed_from)

    @hybrid_property
    def submissions_allowed_to(self):
        return time.strip_seconds(self._submissions_allowed_to)

    @hybrid_property
    def archive_from(self):
        return time.strip_seconds(self._archive_from)

    @submissions_allowed_from.setter
    def submissions_allowed_from(self, submissions_allowed_from):
        if self._submissions_allowed_to is not None and submissions_allowed_from > self._submissions_allowed_to:
            raise PortalDbError()
        if self._archive_from is not None and submissions_allowed_from > self._archive_from:
            raise PortalDbError()
        self._submissions_allowed_from = time.strip_seconds(submissions_allowed_from)

    @submissions_allowed_to.setter
    def submissions_allowed_to(self, submissions_allowed_to):
        if self._submissions_allowed_from is not None and self._submissions_allowed_from > submissions_allowed_to:
            raise PortalDbError()
        if self._archive_from is not None and self._archive_from < submissions_allowed_to:
            raise PortalDbError()
        self._submissions_allowed_to = time.strip_seconds(submissions_allowed_to)

    @archive_from.setter
    def archive_from(self, archive_from):
        if self._submissions_allowed_from is not None and self._submissions_allowed_from > archive_from:
            raise PortalDbError()
        if self._submissions_allowed_to is not None and self._submissions_allowed_to > archive_from:
            raise PortalDbError()
        self._archive_from = time.strip_seconds(archive_from)

    def __init__(self, project, test_files_source) -> None:
        self.project = project
        self.test_files_source = test_files_source

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Role(db.Model, EntityBase):
    __tablename__ = 'role'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(50), nullable=False)
    description = db.Column(db.Text)
    users = db.relationship("User", secondary="users_roles")
    course_id = db.Column(db.Integer, db.ForeignKey('course.id'), nullable=False)
    course = db.relationship("Course", back_populates="roles", uselist=False)

    permissions = db.relationship("RolePermissions", back_populates='role', cascade="all, delete-orphan",
                                  passive_deletes=True, lazy='joined', uselist=False)

    def set_permissions(self, **kwargs):
        for k, w in kwargs.items():
            if hasattr(self.permissions, k) and k not in ('id', 'role_id', 'role'):
                setattr(self.permissions, k, w)

    __table_args__ = (
        db.UniqueConstraint('course_id', 'name', name='course_unique_name'),
    )

    def __init__(self, course, name):
        self.course = course
        self.name = name
        self.permissions = RolePermissions(self)

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class RolePermissions(db.Model, EntityBase):
    __tablename__ = 'rolePermissions'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    role_id = db.Column(db.Integer, db.ForeignKey('role.id'), nullable=False)
    role = db.relationship("Role", back_populates="permissions", uselist=False)

    # all default to False, explicit setting via role.set_permissions is required
    viewCourseLimited = db.Column(db.Boolean, default=False, nullable=False)
    viewCourseFull = db.Column(db.Boolean, default=False, nullable=False)
    updateCourse = db.Column(db.Boolean, default=False, nullable=False)
    setNotesAccessToken = db.Column(db.Boolean, default=False, nullable=False)
    assignRoles = db.Column(db.Boolean, default=False, nullable=False)
    writeRole = db.Column(db.Boolean, default=False, nullable=False)
    readRole = db.Column(db.Boolean, default=False, nullable=False)
    writeProject = db.Column(db.Boolean, default=False, nullable=False)
    deleteProject = db.Column(db.Boolean, default=False, nullable=False)
    archiveProject = db.Column(db.Boolean, default=False, nullable=False)
    writeGroup = db.Column(db.Boolean, default=False, nullable=False)
    readGroup = db.Column(db.Boolean, default=False, nullable=False)
    deleteGroup = db.Column(db.Boolean, default=False, nullable=False)
    readAllSubmissions = db.Column(db.Boolean, default=False, nullable=False)
    readOwnSubmissions = db.Column(db.Boolean, default=False, nullable=False)
    readGroupSubmissions = db.Column(db.Boolean, default=False, nullable=False)  # OPTIONAL
    viewAllSubmissionFiles = db.Column(db.Boolean, default=False, nullable=False)
    createSubmission = db.Column(db.Boolean, default=False, nullable=False)
    resubmitSubmission = db.Column(db.Boolean, default=False, nullable=False)
    readAllReviews = db.Column(db.Boolean, default=False, nullable=False)
    readOwnReviews = db.Column(db.Boolean, default=False, nullable=False)
    writeOwnReview = db.Column(db.Boolean, default=False, nullable=False)
    writeAllReview = db.Column(db.Boolean, default=False, nullable=False)
    evaluateSubmission = db.Column(db.Boolean, default=False, nullable=False)

    def __init__(self, role):
        self.role = role

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Group(db.Model, EntityBase):
    __tablename__ = 'group'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(20), nullable=False)
    course_id = db.Column(db.Integer, db.ForeignKey('course.id'), nullable=False)
    course = db.relationship("Course", back_populates="groups", uselist=False)
    users = db.relationship("User", secondary="users_groups")

    __table_args__ = (
        db.UniqueConstraint('course_id', 'name', name='course_unique_name'),
    )

    def __init__(self, course, name) -> None:
        self.course = course
        self.name = name

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class SubmissionState(enum.Enum):
    CREATED = 1
    READY = 2
    QUEUED = 3
    IN_PROGRESS = 4
    FINISHED = 5


class Submission(db.Model, EntityBase):
    __tablename__ = 'submission'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    processing_scheduled_for = db.Column(db.TIMESTAMP(timezone=True))
    parameters = db.Column(db.Text, nullable=False)
    state = db.Column(db.Enum(SubmissionState, name='SubmissionState'), nullable=False)
    note = db.Column(db.Text)

    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)
    user = db.relationship("User", back_populates="submissions", uselist=False)
    project_id = db.Column(db.Integer, db.ForeignKey('project.id'), nullable=False)
    project = db.relationship("Project", uselist=False)
    review = db.relationship("Review", back_populates="submission", cascade="all, delete-orphan",
                             passive_deletes=True, uselist=False)

    def change_state(self, new_state):
        # open to extension (state transition validation, ...)
        self.state = new_state

    def __init__(self, user, project, parameters, review=None):
        self.user = user
        self.project = project
        self.review = review
        self.parameters = parameters
        self.state = SubmissionState.CREATED

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Review(db.Model, EntityBase):
    __tablename__ = 'review'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    submission_id = db.Column(db.Integer, db.ForeignKey('submission.id'), nullable=False)
    submission = db.relationship("Submission", uselist=False, back_populates="review")
    review_items = db.relationship("ReviewItem", back_populates="review",
                                   cascade="all, delete-orphan", passive_deletes=True)

    def __init__(self, submission):
        self.submission = submission

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class ReviewItem(db.Model, EntityBase):
    __tablename__ = 'reviewItem'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    review_id = db.Column(db.Integer, db.ForeignKey('review.id'), nullable=False)
    review = db.relationship("Review", uselist=False, back_populates="review_items")
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'))
    user = db.relationship("User", uselist=False, back_populates="review_items")  # the review item's author
    content = db.Column(db.Text)
    file = db.Column(db.String(100), nullable=False)
    line = db.Column(db.Integer, nullable=False)

    def __init__(self, user, review, file, line):
        self.review = review
        self.user = user
        self.file = file
        self.line = line

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Component(db.Model, EntityBase):
    __tablename__ = 'component'
    id = db.Column(db.Text(length=36), default=lambda: str(uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(30), nullable=False, unique=True)
    access_token = db.Column(db.String(100))
    ip_address = db.Column(db.String(50))


# source: http://docs.sqlalchemy.org/en/latest/orm/events.html
@event.listens_for(Submission.user_id, 'set', named=True)
def receive_set(**kw):
    submission = kw['target']
    value = kw['value']

    if submission.user_id is not None and submission.user_id != value:
        raise PortalDbError()


# Associations - must come after definitions of the affected classes
users_groups = db.Table("users_groups", db.Model.metadata,
                        db.Column("user_id", db.Integer, db.ForeignKey('user.id')),
                        db.Column("group_id", db.Integer, db.ForeignKey('group.id')))

users_roles = db.Table("users_roles", db.Model.metadata,
                       db.Column("user_id", db.Integer, db.ForeignKey('user.id')),
                       db.Column("role_id", db.Integer, db.ForeignKey('role.id')))

projects_groups = db.Table("projects_groups", db.Model.metadata,
                           db.Column("project_id", db.Integer, db.ForeignKey('project.id')),
                           db.Column("group_id", db.Integer, db.ForeignKey('group.id')))

User.roles = db.relationship("Role", secondary="users_roles")
User.groups = db.relationship("Group", secondary="users_groups")
Project.groups = db.relationship("Group", secondary="projects_groups")
