from marshmallow import Schema, fields, validates_schema, ValidationError
from marshmallow_enum import EnumField

from portal.database.models import SubmissionState


class UserSchema(Schema):
    id = fields.Str(dump_only=True)
    uco = fields.Int()
    email = fields.Email()
    username = fields.Str()
    name = fields.Str()
    is_admin = fields.Bool()
    submissions = fields.Nested('SubmissionSchema', only=('id', 'note', 'state'), many=True)
    review_items = fields.Nested('ReviewItemSchema', only=('id',), many=True)  # TODO: what to show


class CourseSchema(Schema):
    id = fields.Str(dump_only=True)
    name = fields.Str()
    codename = fields.Str()
    roles = fields.Nested('RoleSchema', only=('id', 'name', 'description'), many=True)
    groups = fields.Nested('GroupSchema', only=('id', 'name'), many=True)
    projects = fields.Nested('ProjectSchema', only=('id', 'name'), many=True)


class SubmissionFileParamsSchema(Schema):
    source = fields.Nested('SubmissionFileSourceSchema')


class SubmissionFileSourceSchema(Schema):
    type = fields.Str()
    url = fields.Str()  # git@gitlab.fi.muni.cz:foo/project.git
    branch = fields.Str()
    checkout = fields.Str()

    @validates_schema
    def validate_present_fields(self, data):
        if data.get('type') is None:
            raise ValidationError(f"Submission source type is required. Supported values: 'git', 'zip'.")
        if data['type'] not in ('git', 'zip'):
            raise ValidationError(f"Submission source unsupported: {data['type']}. Supported sources: git, zip. ")
        if data['type'] == 'git':
            if any(('url', 'branch', 'checkout')) not in data.keys:
                raise ValidationError(f"File source 'git' requires precisely 'url', 'branch' and 'checkout' attributes.")
        if data['type'] == 'zip':
            if any(('url', 'branch', 'checkout')) in data.keys:
                raise ValidationError(f"File source 'zip' does not require any other attributes.")


class SubmissionCreateSchema(Schema):
    file_params = fields.Nested('SubmissionFileParamsSchema')
    project_params = fields.Dict()


class ProjectSchema(Schema):
    # TODO: add state as enum
    id = fields.Str(dump_only=True)
    name = fields.Str()
    config = fields.Nested('ProjectConfigSchema', exclude=('id', '_submissions_allowed_from', '_submissions_allowed_to',
                                                           '_archive_from'))  # TODO: check hybrid property behaviour
    course = fields.Nested('CourseSchema', only=('id', 'name', 'codename'))
    submissions = fields.Nested('SubmissionSchema', only=('id', 'uco', 'email', 'username'), many=True)


class ProjectConfigSchema(Schema):
    id = fields.Str(dump_only=True)
    project = fields.Nested('ProjectSchema', only=('id', 'name'))
    test_files_source = fields.Str()
    file_whitelist = fields.Str()
    pre_submit_script = fields.Str()
    post_submit_script = fields.Str()
    submission_parameters = fields.Str()
    submission_scheduler_config = fields.Str()
    submissions_allowed_from = fields.LocalDateTime()  # or just DateTime?
    submissions_allowed_to = fields.LocalDateTime()
    archive_from = fields.LocalDateTime()


class RoleSchema(Schema):
    id = fields.Str(dump_only=True)
    name = fields.Str()
    description = fields.Str()
    users = fields.Nested('UserSchema', only=('id', 'uco', 'email', 'username'), many=True)
    course = fields.Nested('CourseSchema', only=('id', 'name', 'codename'))
    # this could be done prettier (e.g. flattening the structure, showing only set fields)
    permissions = fields.Nested('RolePermissionsSchema', exclude=('id', 'role'))


class RolePermissionsSchema(Schema):
    id = fields.Str(dump_only=True)
    role = fields.Nested('RoleSchema', only=('id', 'name', 'description'))
    viewCourseLimited = fields.Bool()
    viewCourseFull = fields.Bool()
    updateCourse = fields.Bool()
    setNotesAccessToken = fields.Bool()
    assignRoles = fields.Bool()
    writeRole = fields.Bool()
    readRole = fields.Bool()
    writeProject = fields.Bool()
    deleteProject = fields.Bool()
    archiveProject = fields.Bool()
    writeGroup = fields.Bool()
    readGroup = fields.Bool()
    deleteGroup = fields.Bool()
    readAllSubmissions = fields.Bool()
    readOwnSubmissions = fields.Bool()
    readGroupSubmissions = fields.Bool()
    viewAllSubmissionFiles = fields.Bool()
    createSubmission = fields.Bool()
    resubmitSubmission = fields.Bool()
    readAllReviews = fields.Bool()
    readOwnReviews = fields.Bool()
    writeOwnReview = fields.Bool()
    writeAllReview = fields.Bool()
    evaluateSubmission = fields.Bool()


class GroupSchema(Schema):
    id = fields.Str(dump_only=True)
    name = fields.Str()
    course = fields.Nested('CourseSchema')
    users = fields.Nested('UserSchema', only=('id', 'uco', 'email', 'username'), many=True)


class SubmissionSchema(Schema):
    id = fields.Str(dump_only=True)
    note = fields.Str()
    state = EnumField(SubmissionState)
    scheduled_for = fields.LocalDateTime()
    parameters = fields.Dict()
    project = fields.Nested('ProjectSchema', only=('id', 'name', 'course.id'))  # check dot notation
    user = fields.Str()  # id


class ReviewSchema(Schema):
    id = fields.Str(dump_only=True)
    submission = fields.Nested('SubmissionSchema', only=('id', 'note', 'state'))
    review_items = fields.Nested('ReviewItemSchema', many=True)


class ReviewItemSchema(Schema):
    id = fields.Str(dump_only=True)
    content = fields.Str()
    file = fields.Str()
    line = fields.Int()
    review = fields.Nested('ReviewSchema', only=('id',))
    user = fields.Nested('UserSchema', only=('id', 'uco', 'email', 'username'))


class ComponentSchema(Schema):
    id = fields.Str(dump_only=True)
    name = fields.Str()
    access_token = fields.Str()
    ip_address = fields.Str()

