from flask import Blueprint, request
from flask_restful import Api, Resource, abort
from marshmallow import ValidationError

from portal.database.models import Course
from portal.rest import CourseSchema
from portal.service import service
from portal.tools.decorators import error_handler
from portal.tools.logging import log


courses = Blueprint('courses', __name__, url_prefix='/courses')
courses_api = Api(courses)


course_schema = CourseSchema()
courses_schema = CourseSchema(many=True, only=('id', 'name', 'codename'))


class CourseResource(Resource):
    @error_handler
    def get(self, cid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not find course {cid}.")
        return course_schema.dump(course)

    @error_handler
    def delete(self, cid):
        log.info(f"DELETE to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not find course {cid}.")
        service.delete_entity(course)
        return '', 204

    @error_handler
    def put(self, cid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for course update.')
        try:
            data = course_schema.load(json_data)
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        course.name = data['name']
        course.codename = data['codename']
        service.write_entity(course)
        log.debug(f"Updated course {cid} to {course}.")
        return '', 204


class CourseList(Resource):
    @error_handler
    def get(self):
        log.info(f"GET to {request.url}")
        courses = Course.query.all()
        return courses_schema.dump(courses)

    @error_handler
    def post(self):
        log.info(f"POST to {request.url}")
        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for course creation.')

        try:
            data = course_schema.load(json_data)
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors
        name = data[0]['name']
        codename = data[0]['codename']
        new_course = Course(name=name, codename=codename)

        service.write_entity(new_course)
        log.debug(f"Created course={new_course}")
        return course_schema.dump(new_course)[0], 201


class CourseNotesToken(Resource):

    @error_handler
    def get(self, cid):
        course = service.find_course(cid)
        return course.notes_access_token

    @error_handler
    def put(self, cid):
        course = service.find_course(cid)
        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for notes token update.')

        course.notes_access_token = json_data  # TODO: validation
        service.write_entity(course)
        log.info(f"Updated notes access token for course id={cid} to {json_data}")
        return '', 204


courses_api.add_resource(CourseResource, '/<string:cid>')
courses_api.add_resource(CourseList, '/')
courses_api.add_resource(CourseNotesToken, "/<string:cid>/notes_access_token")
