import json

from flask import Blueprint, request
from flask_restful import Api, Resource, abort
from marshmallow import ValidationError
from flask_jwt_extended import JWTManager, jwt_required, create_access_token, get_jwt_identity

from portal.rest import ProjectSchema, ProjectConfigSchema, SubmissionSchema, SubmissionCreateSchema
from portal.tools.decorators import error_handler
from portal.tools.logging import log
from portal.database.models import Project
from portal.service import service

projects = Blueprint('projects', __name__, url_prefix='/courses/<string:cid>/projects')
projects_api = Api(projects)

project_schema = ProjectSchema()
projects_schema = ProjectSchema(many=True, only=('id', 'name'))
config_schema = ProjectConfigSchema()
# add user information to submission schema? How?
submissions_schema = SubmissionSchema(many=True, only=('id', 'parameters', 'state', 'note'))
submission_schema = SubmissionSchema()


class ProjectResource(Resource):
    @error_handler
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        return project_schema.dump(project)

    @error_handler
    def delete(self, cid, pid):
        log.info(f"DELETE to {request.url}")
        course = service.find_course(cid)
        project = service.find_project(course, pid)

        service.delete_entity(project)
        return '', 204

    @error_handler
    def put(self, cid, pid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)

        project = service.find_project(course, pid)

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for project update.')
        try:
            data = project_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        project.name = data['name']
        service.write_entity(project)
        log.debug(f"Updated project {pid} to {project}.")
        return '', 204


class ProjectsList(Resource):
    @error_handler
    def get(self, cid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        return projects_schema.dump(course.projects)

    @error_handler
    def post(self, cid):
        log.info(f"POST to {request.url}")
        course = service.find_course(cid)
        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for project creation.')

        try:
            data = project_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        new_project = Project(course=course, name=data['name'],
                              test_files_source=data['test_files_source'])
        service.write_entity(new_project)
        log.debug(f"Created project={new_project} for course {cid}")
        return project_schema.dump(new_project)[0], 201


class ProjectConfigResource(Resource):
    @error_handler
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        project = service.find_project(course, pid)

        return config_schema.dump(project.config)

    @error_handler
    def put(self, cid, pid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)
        project = service.find_project(course, pid)

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for project configuration update.')
        try:
            data = config_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        project.set_config(**data)
        log.debug(f"Updated configuration for project {pid} in course {cid} to {data}.")
        return '', 204


class ProjectSubmissions(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # TODO: user=<user_selector>` - filters submissions by user
        return submissions_schema.dump(project.submissions)

    @error_handler
    @jwt_required
    def post(self, cid, pid):
        create_schema = SubmissionCreateSchema()
        log.info(f"POST to {request.url}")
        # TODO FIX: do not use find_user but autorize_user
        user = service.find_user(get_jwt_identity())
        if not user:
            abort(401, message="Invalid access token: user not found.")
        # authorization here

        course = service.find_course(cid)
        project = service.find_project(course, pid)

        # check if a new submission can be created by this user in this project
        # TODO: configure conflict_timeout in project or remove completely (Kontr handles its own queue)
        if not service.can_create_submission(user, project):
            # maybe use 409, 403
            abort(429,
                  message=f"New submission for user {user.username} denied: a submission for this user "
                          f"is created or queued.")

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for submission creation.')
        try:
            data = create_schema.load(json_data, partial=True)
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        project_params = data.get(
            'project_params')  # data for Kontr processing, of no use to portal or submission q
        if project_params:
            project_params_string = json.dumps(project_params)  # not pretty, but should do for now
        else:
            project_params_string = ""
        new_submission = service.create_submission(user=user, project=project,
                                                   file_params=data['file_params'],
                                                   project_params_string=project_params_string)

        return submission_schema.dump(new_submission)


class ProjectTestFiles(Resource):
    @error_handler
    def get(self, cid, pid):
        # Gets test files (feature)
        return "Not implemented."


projects_api.add_resource(ProjectResource, '/<string:pid>')
projects_api.add_resource(ProjectsList, '/')
projects_api.add_resource(ProjectConfigResource, '/<string:pid>/config')
projects_api.add_resource(ProjectSubmissions, '/<string:pid>/submissions')
projects_api.add_resource(ProjectTestFiles, '/<string:pid>/files')
