from flask import Blueprint, request
from flask_restful import Api, Resource, abort
from marshmallow import ValidationError

from portal.database.models import Role
from portal.rest import RoleSchema, RolePermissionsSchema, UserSchema
from portal.service import service
from portal.tools.decorators import error_handler
from portal.tools.logging import log


roles = Blueprint('roles', __name__, url_prefix='/courses/<string:cid>/roles')
roles_api = Api(roles)

role_schema = RoleSchema()
roles_schema = RoleSchema(many=True, only=('id', 'name', 'description'))
permissions_schema = RolePermissionsSchema()
users_schema = UserSchema(many=True, only=('id', 'uco', 'email', 'username'))


class RoleResource(Resource):
    @error_handler
    def get(self, cid, rid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        return role_schema.dump(role)

    @error_handler
    def delete(self, cid, rid):
        log.info(f"DELETE to {request.url}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        service.delete_entity(role)
        return '', 204

    @error_handler
    def put(self, cid, rid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)

        role = service.find_role(course, rid)
        if not role:
            abort(404, message=f"Could not update role {rid} in course {cid}: role not found.")

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for role update.')
        try:
            data = role_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        role.description = data['description']
        role.name = data['name']
        # updating course?
        service.write_entity(role)
        log.debug(f"Updated role {rid} to {role}.")
        return '', 204


class RoleList(Resource):

    @error_handler
    def get(self, cid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        return roles_schema.dump(course.roles)

    @error_handler
    def post(self, cid):
        log.info(f"POST to {request.url}")
        course = service.find_course(cid)
        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for role creation.')
        try:
            data = role_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        new_role = Role(course=course, name=data['name'])
        new_role.description = data.get('description') or ""
        service.write_entity(new_role)
        log.debug(f"Created role={new_role} for course {cid}")
        return role_schema.dump(new_role)[0], 201


class RolePermissions(Resource):
    @error_handler
    def get(self, cid, rid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        return roles_schema.dump(role.permissions)

    @error_handler
    def put(self, cid, rid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for update of role permissions.')
        try:
            data = role_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        role.set_permissions(**data)
        log.debug(f"Updated permissions for role {rid} in course {cid} to {data}.")
        return '', 204


class RoleUsersList(Resource):
    @error_handler
    def get(self, cid, rid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)

        return users_schema.dump(role.users)

    @error_handler
    def put(self, cid, rid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        json_data = request.get_json()
        try:
            data = users_schema.load(json_data)[0]  # presumes the json contains full entities; change to ids?
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        # everything from users_add is added, THEN everything from users_remove is subtracted
        users_old = set(role.users)
        users_add = data.get('add')
        users_remove = data.get('remove')
        if users_add is not None:
            users_old.union(users_add)
        if users_remove is not None:
            users_old.difference(users_remove)

        log.info(f"Updated users of role {rid} to {data}.")
        return '', 204


class RoleUser(Resource):
    @error_handler
    def put(self, cid, rid, uid):
        log.info(f"PUT to {request.url}")
        log.debug(f"ADD User to role: course_id={cid}, role_id={rid}, user_id={uid}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        user = service.find_user(uid)

        if user not in role.users:
            role.users.append(user)
            log.info(f"Added user {uid} to role {rid} in course {cid}.")
        else:
            log.info(f"User {uid} is already in role {rid} in course {cid}: no change.")
        return '', 204

    @error_handler
    def delete(self, cid, rid, uid):
        log.info(f"DELETE to {request.url}")
        log.debug(f"REMOVE User from role: course_id={cid}, role_id={rid}, user_id={uid}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        user = service.find_user(uid)

        try:
            role.users.remove(user)
        except ValueError:
            abort(400, message=f"Could not remove user {uid} from role {rid} in course {cid}: "
                               f"role does not contain user.")
        log.info(f"Removed user {uid} from role {rid} in course {cid}.")
        return '', 204


roles_api.add_resource(RoleResource, "/<string:rid>")
roles_api.add_resource(RoleList, "/")
roles_api.add_resource(RolePermissions, "/<string:rid>/permissions")
roles_api.add_resource(RoleUsersList, "/<string:rid>/users")
roles_api.add_resource(RoleUser, "/<string:rid>/users/<string:uid>")

