from flask import Blueprint, request
from flask_restful import Api, Resource, abort
from marshmallow import ValidationError

from portal.rest import SubmissionSchema
from portal.tools.decorators import error_handler
from portal.tools.logging import log
from portal.service import service

submissions = Blueprint('submissions', __name__, url_prefix='/submissions')
submissions_api = Api(submissions)

submission_schema = SubmissionSchema()
submissions_schema = SubmissionSchema(many=True, only=('id', ))


class SubmissionResource(Resource):
    @error_handler
    def get(self, sid):
        log.info(f"GET to {request.url}")
        submission = service.find_submission(sid)
        # TODO: add stuff to the schema, needs testing
        return submission_schema.dump(submission)


class SubmissionState(Resource):
    state_schema = SubmissionSchema(only=('state',))

    @error_handler
    def get(self, sid):
        log.info(f"GET to {request.url}")
        submission = service.find_submission(sid)
        # returning the enum value as a string should also be okay
        return self.state_schema.dump(submission.state)

    @error_handler
    def put(self, sid):
        log.info(f"PUT to {request.url}")
        submission = service.find_submission(sid)

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for submission state update.')
        try:
            data = self.state_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        submission.change_state(data['state'])
        service.write_entity(submission)
        log.info(f"Updated state of submission {sid} to {submission.state}")
        return '', 204


class SubmissionResult(Resource):
    @error_handler
    def put(self, sid):
        log.info(f"PUT to {request.url}")
        submission = service.find_submission(sid)

        # TODO load zip, save to storage


class SubmissionFiles(Resource):
    @error_handler
    def get(self, sid):
        # the user wants a zip or a json

        pass


class SubmissionSourceFiles(Resource):
    @error_handler
    def get(self, sid):
        pass


class SubmissionResultFiles(Resource):
    @error_handler
    def get(self, sid):
        pass


class SubmissionTestFiles(Resource):
    @error_handler
    def get(self, sid):
        pass


class SubmissionSources(Resource):
    @error_handler
    def get(self, sid):
        pass


class SubmissionResubmit(Resource):
    @error_handler
    def post(self, sid):
        log.info(f"POST to {request.url}")
        source_submission = service.find_submission(sid)
        # create new submission by copying files from the source submission in storage
        # TODO: add to service




'''
processing url parameters: https://stackoverflow.com/questions/24892035/python-flask-how-to-get-parameters-from-a-url
from flask import request

@app.route('/my-route')
def my_route():
  page = request.args.get('page', default = 1, type = int)
  filter = request.args.get('filter', default = '*', type = str)
'''

submissions_api.add_resource(SubmissionResource, '/<string:sid>')  # needs extension
submissions_api.add_resource(SubmissionState, '/<string:sid>/state')  # OK
submissions_api.add_resource(SubmissionResult, '/<string:sid>/result')
submissions_api.add_resource(SubmissionResubmit, '/<string:sid>/resubmit')

submissions_api.add_resource(SubmissionFiles, '/<string:sid>/files')
submissions_api.add_resource(SubmissionSourceFiles, '/<string:sid>/files/sources')
submissions_api.add_resource(SubmissionResultFiles, '/<string:sid>/files/results')

submissions_api.add_resource(SubmissionTestFiles, '/<string:sid>/test_files')
submissions_api.add_resource(SubmissionSources, '/<string:sid>/sources')
