from flask import Blueprint, request
from flask_restful import Api, Resource, abort
from marshmallow import ValidationError

from portal.rest import UserSchema
from portal.service.service import delete_entity, write_entity, find_user
from portal.database.models import User
from portal.tools.logging import log

users = Blueprint('users', __name__, url_prefix='/users')
users_api = Api(users)

user_schema = UserSchema()
users_schema = UserSchema(many=True, only=('id', 'username', 'uco', 'email'))


class UserResource(Resource):
    def get(self, uid):
        log.info(f"GET to {request.url}")
        user = find_user(uid)
        if not user:
            abort(404, message=f"Could not find user {uid}.")
        return user_schema.dump(user)

    def put(self, uid):
        log.info(f"PUT to {request.url}")
        user = find_user(uid)
        if not user:
            abort(404, message=f"Could not find user {uid}.")

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for user update.')
        try:
            data = user_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        user.name = data['name']
        user.email = data['email']
        user.uco = data['uco']
        user.username = data['username']
        user.is_admin = data['is_admin']
        # TODO all; care for password
        write_entity(user)
        log.debug(f"Updated user {uid} to {user}.")
        return '', 204

    def delete(self, uid):
        log.info(f"DELETE to {request.url}")
        user = find_user(uid)
        if not user:
            abort(404, message=f"Could not find user {uid}.")
        delete_entity(user)
        return '', 204


class UserList(Resource):
    def get(self):
        log.info(f"GET to {request.url}")
        users = User.query.all()
        return users_schema.dump(users)

    def post(self):
        log.info(f"POST to {request.url}")
        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for user creation.')
        try:
            data = user_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        admin = data.get('is_admin') or False
        new_user = User(uco=data['uco'], email=data['email'], username=data['username'], is_admin=admin)
        write_entity(new_user)
        log.debug(f"Created user={User}")
        return user_schema.dump(new_user)[0], 201
    

users_api.add_resource(UserResource, '/<string:uid>')
users_api.add_resource(UserList, '/')

