from portal.rest import UserSchema
from portal.database.models import User
import json

''' tests for user API
* [GET] /users/ : list all users (only sysadmin)
	- Query:
		- group=gid : filter by group (teacher)
		- course=cid : filter by course (teacher)
* [POST] /users/ : create a new user (only sysadmin (component - IS import?))

* [GET] /users/<uid> : view info about one user
	- if uid == auth.current_user: allow, else only for sysadmin; same for derived paths
* [PUT] /users/<uid> : update user; different shcema validations/format
	- non-admin: name, password_hash, only self
	- admin : + uco, email, username, is_admin, for all
* [DELETE] /users/<uid> : delete a user from the database
	- only sysadmin
	
* [GET] /users/<uid>/submissions : list all submissions (paginated); needs submission id, project id, course id; (maybe timing info)
	- **Query:**
		- project=pid : filter by project
		- course=cid : filter by course, group by course projects
	- clickable, redirects to /submissions/<id>

* [GET] /users/<uid>/roles : list user's roles 
	- same helper function as authorization
	- maybe inaccessible through UI? (what for?)
	- filter by course

* [GET] /users/<uid>/courses : list courses the user has at least one role in (is part of)
	- show only recent/active (non-archived)?

* [GET] /users/<uid>/groups : list groups the user is a part of
	- group, filter by course
	
* [GET] /user/<uid>/reviews : list reviews associated with the user
	- reviews the user contributed to
	- reviews associated with the user's submissions: accessible through the user's submissions

'''
user_schema = UserSchema()
users_schema = UserSchema(many=True, only=('id', 'uco', 'username', 'email'))


def assert_user(expected: User, actual: dict):
    assert expected.id == actual['id']
    assert expected.uco == actual['uco']
    assert expected.email == actual['email']
    assert expected.username == actual['username']
    assert expected.name == actual['name']
    assert expected.is_admin == actual['is_admin']


def compare_user(expected: User, actual:dict):
    return expected.id == actual['id'] and expected.uco == actual['uco'] and expected.email == actual['email']


def assert_user_in(user_list: list, user: dict):
    for u in user_list:
        if compare_user(u, user):
            return True
    return False


def test_create(client):
    request_dict = {
        "uco": 445574,
        "email": "baz@foo.com",
        "username": "xuser",
    }
    request_json = json.dumps(request_dict)
    response = client.post('/users/', data=request_json, headers={"content-type": "application/json"})
    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    resp_new_user = json.loads(str(response.get_data().decode("utf-8")))
    db_users = User.query.all()
    assert len(db_users) == 3
    assert_user_in(db_users, resp_new_user)


#follow_redirects=True
def test_list(client):
    response = client.get('/users/')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'
    data = str(response.get_data().decode("utf-8"))
    users = json.loads(data)
    assert len(users) == 2
    db_users = User.query.all()
    assert_user_in(db_users, users[0])


def test_read(client):
    user = User.query.all()[0]
    response = client.get(f'/users/{user.id}')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    data = json.loads(str(response.get_data().decode("utf-8")))
    assert_user(user, data)


def test_update(client):
    user = User.query.all()[0]
    request_dict = {
        "name": "new_name",
        "email": "new_email@domain.net",
        "uco": 456789,
        "username": "xnew",
        "is_admin": True,
    }
    request_json = json.dumps(request_dict)
    response = client.put(f'/users/{user.id}', data=request_json, headers={"content-type": "application/json"})

    assert response.status_code == 204
    assert response.mimetype == 'application/json'
    updated_entity = User.query.filter_by(id=user.id).first()
    assert updated_entity.name == request_dict["name"]
    assert updated_entity.email == request_dict["email"]
    assert updated_entity.uco == request_dict["uco"]
    assert updated_entity.username == request_dict["username"]
    assert updated_entity.is_admin == request_dict["is_admin"]
    assert len(User.query.all()) == 2


def test_delete(client):
    user = User.query.all()[0]

    response = client.delete(f"/users/{user.id}")
    assert response.status_code == 204
    assert response.mimetype == 'application/json'

    db_users = User.query.all()
    assert len(db_users) == 1
    assert user not in db_users
